"""
Alternative exception formatter handling __traceback_supplement__ and
__traceback_info__ if present.

Implements portions of the 'traceback' standard library module and the
ITracebackSupplement interface of Zope3. Also see interfaces.py.

$Id: exceptionformatter.py,v 1.12 2005/10/30 09:02:18 stefan Exp $
"""

import sys
import linecache

class ExceptionFormatter(object):
    limit = getattr(sys, 'tracebacklimit', None)
    debug = True

    def format_exception(self, e, v, tb, limit=None):
        """Formats a stack trace and the exception information."""
        if tb:
            list = ['Traceback (most recent call last):\n']
            list.extend(self.format_tb(tb, limit))
        else:
            list = []
        list.append(self.format_exception_only(e, v))
        return list

    def format_tb(self, tb, limit=None):
        """A shorthand for 'format_list(extract_tb(tb, limit))'."""
        return self.format_list(self.extract_tb(tb, limit))

    def extract_tb(self, tb, limit=None):
        """Returns a list of up to limit pre-processed entries from traceback."""
        if limit is None:
            limit = self.limit
        list = []
        n = 0
        while tb is not None and (limit is None or n < limit):
            f = tb.tb_frame
            lineno = tb.tb_lineno
            co = f.f_code
            filename = co.co_filename
            name = co.co_name
            locals = f.f_locals
            globals = f.f_globals

            linecache.checkcache(filename)
            line = linecache.getline(filename, lineno)
            if line:
                line = line.strip()
            else:
                line = None

            tbs = locals.get('__traceback_supplement__')
            if tbs is None:
                tbs = globals.get('__traceback_supplement__')
            if tbs is not None:
                try:
                    factory, args = tbs[0], tbs[1:]
                    tbs = factory(*args)
                except:
                    tbs = None
                    if self.debug:
                        import traceback
                        traceback.print_exc()

            tbi = locals.get('__traceback_info__')
            if tbi is None:
                tbi = globals.get('__traceback_info__')

            list.append((filename, lineno, name, line, tbs, tbi))
            tb = tb.tb_next
            n = n + 1
        return list

    def format_list(self, extracted_list):
        """Formats a list of traceback entry tuples for printing."""
        list = []
        for filename, lineno, name, line, tbs, tbi in extracted_list:
            item = '  File "%s", line %d, in %s\n' % (filename, lineno, name)
            if line:
                item += '    %s\n' % line.strip()

            if tbs is not None:
                supplement = self.format_traceback_supplement(tbs)
                if supplement:
                    item += '%s\n' % supplement

            if tbi is not None:
                item += '%s\n' % self.format_traceback_info(tbi)

            list.append(item)
        return list

    def format_traceback_supplement(self, supplement):
        """Formats a traceback supplement as string."""
        lines = []
        append = lines.append
        format_supplement_line = self.format_supplement_line

        source_url = getattr(supplement, 'source_url', None)
        if source_url:
            append(format_supplement_line(source_url))

        line = getattr(supplement, 'line', 0)
        column = getattr(supplement, 'column', -1)
        if line is not None and line > 0:
            if column is not None and column >= 0:
                append(format_supplement_line('Line %s, Column %s' % (line, column)))
            else:
                append(format_supplement_line('Line %s' % line))
        elif column is not None and column >= 0:
            append(format_supplement_line('Column %s' % column))

        expression = getattr(supplement, 'expression', None)
        if expression:
            append(format_supplement_line('Expression: %s' % expression))

        warnings = getattr(supplement, 'warnings', None)
        if warnings:
            for warning in warnings:
                append(format_supplement_line('Warning: %s' % warning))

        getInfo = getattr(supplement, 'getInfo', None)
        if getInfo is not None:
            try:
                extra = getInfo()
                if extra:
                    append(extra)
            except:
                if self.debug:
                    import traceback
                    traceback.print_exc()

        return '\n'.join(lines)

    def format_supplement_line(self, line):
        """Formats a single line of traceback supplement."""
        return '   - %s' % line

    def format_traceback_info(self, info):
        """Formats a traceback info as string."""
        return self.format_supplement_line('__traceback_info__: %s' % (info,))

    def format_exception_only(self, e, v):
        """Formats the exception part of a traceback."""
        import traceback
        return '\n'.join(traceback.format_exception_only(e, v))

    def format_exc(self, limit=None):
        """Same as 'format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback, limit)'
        but returns a string.
        """
        try:
            e, v, tb = sys.exc_info()
            return ''.join(self.format_exception(e, v, tb, limit))
        finally:
            e = v = tb = None

formatter = ExceptionFormatter()
format_exception = formatter.format_exception
format_exc = formatter.format_exc

