"""
Tests for custom exception formatter

$Id: test_exceptionformatter.py,v 1.5 2005/10/30 17:31:48 stefan Exp $
"""

from unittest import TestCase, TestSuite, makeSuite, main

from django.contrib.exceptionformatter.exceptionformatter import ExceptionFormatter
from django.contrib.exceptionformatter.exceptionformatter import format_exception
from django.contrib.exceptionformatter.exceptionformatter import format_exc


def exc_info():
    import sys
    return sys.exc_info()

def tb():
    return exc_info()[2]

def bomb(exception, tbs=None, tbi=None):
    if tbs is not None:
        __traceback_supplement__ = tbs
    if tbi is not None:
        __traceback_info__ = tbi
    raise exception

class Supplement(object):
    def __init__(self, source_url=None, line=None, column=None, expression=None, warnings=None):
        if source_url is not None:
            self.source_url = source_url
        if line is not None:
            self.line = line
        if column is not None:
            self.column = column
        if expression is not None:
            self.expression = expression
        if warnings is not None:
            self.warnings = warnings

class SupplementX(Supplement):
    def __init__(self, source_url=None, line=None, column=None, expression=None, warnings=None, getInfo=None):
        Supplement.__init__(self, source_url, line, column, expression, warnings)
        self._getInfo = getInfo
    def getInfo(self):
        return self._getInfo

class BrokenSupplement(object):
    def __init__(self):
        raise Exception

__FILE__ = __file__
if __FILE__.endswith('.pyo') or __FILE__.endswith('.pyc'):
    __FILE__ = __FILE__[:-1]


class FormatterTests(TestCase):

    def setUp(self):
        self.fmt = ExceptionFormatter()
        self.fmt.debug = False

    def testExtractTb(self):
        try:
            bomb(Exception)
        except:
            lines = self.fmt.extract_tb(tb())
            self.assertEqual(len(lines), 2)
            self.assertEqual(lines[-1], (__FILE__, 26, 'bomb', 'raise exception', None, None))

    def testExtractTracebackInfo(self):
        try:
            bomb(Exception, tbi='fred')
        except:
            lines = self.fmt.extract_tb(tb())
            self.assertEqual(len(lines), 2)
            self.assertEqual(lines[-1], (__FILE__, 26, 'bomb', 'raise exception', None, 'fred'))

    def testExtractNonStringTracebackInfo(self):
        try:
            bomb(Exception, tbi={'fred': 'barney'})
        except:
            lines = self.fmt.extract_tb(tb())
            self.assertEqual(len(lines), 2)
            self.assertEqual(lines[-1], (__FILE__, 26, 'bomb', 'raise exception', None, {'fred': 'barney'}))

    def testExtractTracebackSupplement(self):
        try:
            bomb(Exception, tbs=(Supplement,))
        except:
            lines = self.fmt.extract_tb(tb())
            self.assertEqual(len(lines), 2)
            self.failUnless(isinstance(lines[-1][-2], Supplement))

    def testTracebackSupplementIsNoneIfUnableToConstruct(self):
        try:
            bomb(Exception, tbs=(BrokenSupplement,))
        except:
            lines = self.fmt.extract_tb(tb())
            self.assertEqual(len(lines), 2)
            self.assertEqual(lines[-1][-2], None)

    def testExtractTracebackWithLimit(self):
        try:
            bomb(Exception)
        except:
            lines = self.fmt.extract_tb(tb(), limit=1)
            self.assertEqual(len(lines), 1)
            self.assertEqual(lines[0], (__FILE__, 105, 'testExtractTracebackWithLimit', 'bomb(Exception)', None, None))

    def testFormatExceptionOnly(self):
        try:
            bomb(AttributeError('barney'))
        except:
            e, v = exc_info()[:2]
            line = self.fmt.format_exception_only(e, v)
            self.assertEqual(line, "AttributeError: barney\n")

    def testFormatExceptionOnlyRepr(self):
        try:
            bomb(AttributeError({'fred': 'barney'}))
        except:
            e, v = exc_info()[:2]
            line = self.fmt.format_exception_only(e, v)
            self.assertEqual(line, "AttributeError: {'fred': 'barney'}\n")

    def testFormatSupplementLine(self):
        # three spaces, a dash, a space, the text
        line = self.fmt.format_supplement_line('barney')
        self.assertEqual(line, '   - barney')

    def testFormatTracebackInfo(self):
        line = self.fmt.format_traceback_info('barney')
        self.assertEqual(line, '   - __traceback_info__: barney')

    def testFormatTracebackInfoRepr(self):
        line = self.fmt.format_traceback_info({'fred': 'barney'})
        self.assertEqual(line, "   - __traceback_info__: {'fred': 'barney'}")

    def testFormatSupplementEmpty(self):
        line = self.fmt.format_traceback_supplement(Supplement())
        self.assertEqual(line, '')

    def testFormatSupplementSourceUrl(self):
        line = self.fmt.format_traceback_supplement(Supplement(source_url='http://bedrock.com'))
        self.assertEqual(line, '   - http://bedrock.com')

    def testExcludeSourceUrlIfEmpty(self):
        line = self.fmt.format_traceback_supplement(Supplement(source_url=''))
        self.assertEqual(line, '')

    def testExcludeSourceUrlIfNone(self):
        line = self.fmt.format_traceback_supplement(Supplement(source_url=None))
        self.assertEqual(line, '')

    def testFormatSupplementLine(self):
        line = self.fmt.format_traceback_supplement(Supplement(line=23))
        self.assertEqual(line, '   - Line 23')

    def testFormatSupplementColumn(self):
        line = self.fmt.format_traceback_supplement(Supplement(column=42))
        self.assertEqual(line, '   - Column 42')

    def testFormatSupplementLineColumn(self):
        line = self.fmt.format_traceback_supplement(Supplement(line=23, column=42))
        self.assertEqual(line, '   - Line 23, Column 42')

    def testExcludeLineIfNone(self):
        line = self.fmt.format_traceback_supplement(Supplement(line=None, column=42))
        self.assertEqual(line, '   - Column 42')

    def testExcludeLineIfZero(self):
        line = self.fmt.format_traceback_supplement(Supplement(line=0, column=42))
        self.assertEqual(line, '   - Column 42')

    def testExcludeLineIfMinusOne(self):
        line = self.fmt.format_traceback_supplement(Supplement(line=-1, column=42))
        self.assertEqual(line, '   - Column 42')

    def testExcludeLineIfMinusTwo(self):
        line = self.fmt.format_traceback_supplement(Supplement(line=-2, column=42))
        self.assertEqual(line, '   - Column 42')

    def testExcludeColumnIfNone(self):
        line = self.fmt.format_traceback_supplement(Supplement(line=23, column=None))
        self.assertEqual(line, '   - Line 23')

    def testIncludeColumnIfZero(self):
        line = self.fmt.format_traceback_supplement(Supplement(line=23, column=0))
        self.assertEqual(line, '   - Line 23, Column 0')

    def testExcludeColumnIfMinusOne(self):
        line = self.fmt.format_traceback_supplement(Supplement(line=23, column=-1))
        self.assertEqual(line, '   - Line 23')

    def testExcludeColumnIfMinusTwo(self):
        line = self.fmt.format_traceback_supplement(Supplement(line=23, column=-2))
        self.assertEqual(line, '   - Line 23')

    def testFormatSupplementExpression(self):
        line = self.fmt.format_traceback_supplement(Supplement(expression='fred'))
        self.assertEqual(line, '   - Expression: fred')

    def testFormatSupplementExpressionRepr(self):
        line = self.fmt.format_traceback_supplement(Supplement(expression=['fred']))
        self.assertEqual(line, "   - Expression: ['fred']")

    def testExcludeExpressionIfEmpty(self):
        line = self.fmt.format_traceback_supplement(Supplement(expression=''))
        self.assertEqual(line, '')

    def testExcludeExpressionIfNone(self):
        line = self.fmt.format_traceback_supplement(Supplement(expression=None))
        self.assertEqual(line, '')

    def testFormatSupplementWarnings(self):
        line = self.fmt.format_traceback_supplement(Supplement(warnings=('fred', 'barney')))
        self.assertEqual(line, '   - Warning: fred\n'
                               '   - Warning: barney')

    def testFormatSupplementWarningsRepr(self):
        line = self.fmt.format_traceback_supplement(Supplement(warnings=(['fred'], ['barney'])))
        self.assertEqual(line, "   - Warning: ['fred']\n"
                               "   - Warning: ['barney']")

    def testExcludeWarningsIfEmpty(self):
        line = self.fmt.format_traceback_supplement(Supplement(warnings=()))
        self.assertEqual(line, '')

    def testExcludeWarningsIfNone(self):
        line = self.fmt.format_traceback_supplement(Supplement(warnings=None))
        self.assertEqual(line, '')

    def testFormatSupplementNewlineSeparated(self):
        supplement = Supplement(source_url='http://bedrock.com',
                                line=23, column=42,
                                expression='fred',
                                warnings=('barney', 'wilma'))
        line = self.fmt.format_traceback_supplement(supplement)
        self.assertEqual(line, '   - http://bedrock.com\n'
                               '   - Line 23, Column 42\n'
                               '   - Expression: fred\n'
                               '   - Warning: barney\n'
                               '   - Warning: wilma')

    def testFormatSupplementGetInfo(self):
        # getInfo is appended as is, without a supplement line prefix
        line = self.fmt.format_traceback_supplement(SupplementX(getInfo='bedrock.com'))
        self.assertEqual(line, 'bedrock.com')

    def testExcludeGetInfoIfEmpty(self):
        line = self.fmt.format_traceback_supplement(SupplementX(getInfo=''))
        self.assertEqual(line, '')

    def testExcludeGetInfoIfNone(self):
        line = self.fmt.format_traceback_supplement(SupplementX(getInfo=None))
        self.assertEqual(line, '')

    def testFormatSupplementGetInfoNewlineSeparated(self):
        supplement = SupplementX(source_url='http://bedrock.com', line=23, column=42,
                                 getInfo='   - Names: pebbles, bambam')
        line = self.fmt.format_traceback_supplement(supplement)
        self.assertEqual(line, '   - http://bedrock.com\n'
                               '   - Line 23, Column 42\n'
                               '   - Names: pebbles, bambam')

    def testFormatList(self):
        tb = ('fred.py', 23, 'get_barney', '', None, None)
        line = self.fmt.format_list([tb])[0]
        self.assertEqual(line, '  File "fred.py", line 23, in get_barney\n')

    def testFormatListWithLine(self):
        tb = ('fred.py', 23, 'get_barney', 'return barneys.get(id)', None, None)
        line = self.fmt.format_list([tb])[0]
        self.assertEqual(line, '  File "fred.py", line 23, in get_barney\n'
                               '    return barneys.get(id)\n')

    def testFormatListWithTracebackInfo(self):
        tb = ('fred.py', 23, 'get_barney', '', None, 'bambam')
        line = self.fmt.format_list([tb])[0]
        self.assertEqual(line, '  File "fred.py", line 23, in get_barney\n'
                               '   - __traceback_info__: bambam\n')

    def testFormatListWithEmptyTracebackInfo(self):
        tb = ('fred.py', 23, 'get_barney', '', None, '')
        line = self.fmt.format_list([tb])[0]
        self.assertEqual(line, '  File "fred.py", line 23, in get_barney\n'
                               '   - __traceback_info__: \n')

    def testFormatListWithTracebackSupplement(self):
        tb = ('fred.py', 23, 'get_barney', '', Supplement('http://bedrock.com', 23, 42), None)
        line = self.fmt.format_list([tb])[0]
        self.assertEqual(line, '  File "fred.py", line 23, in get_barney\n'
                               '   - http://bedrock.com\n'
                               '   - Line 23, Column 42\n')

    def testFormatListWithEmptyTracebackSupplement(self):
        tb = ('fred.py', 23, 'get_barney', '', Supplement(), None)
        line = self.fmt.format_list([tb])[0]
        self.assertEqual(line, '  File "fred.py", line 23, in get_barney\n')

    def testFormatListWithTracebackSupplementGetInfo(self):
        tb = ('fred.py', 23, 'get_barney', '', SupplementX(getInfo='bedrock.com'), None)
        line = self.fmt.format_list([tb])[0]
        self.assertEqual(line, '  File "fred.py", line 23, in get_barney\n'
                               'bedrock.com\n')

    def testFormatListWithLineAndTracebackSupplementAndTracebackInfo(self):
        tb = ('fred.py', 23, 'get_barney', 'return barneys.get(id)',
              Supplement('http://bedrock.com', 23, 42), 'bambam')
        line = self.fmt.format_list([tb])[0]
        self.assertEqual(line, '  File "fred.py", line 23, in get_barney\n'
                               '    return barneys.get(id)\n'
                               '   - http://bedrock.com\n'
                               '   - Line 23, Column 42\n'
                               '   - __traceback_info__: bambam\n')

    def testFormatListMoreThanOneLine(self):
        tb = (('fred.py', 23, 'get_barney', '', None, None),
              ('flintstone.py', 7, 'call_wilma', '', None, None))
        lines = self.fmt.format_list(tb)
        self.assertEqual(lines[0], '  File "fred.py", line 23, in get_barney\n')
        self.assertEqual(lines[1], '  File "flintstone.py", line 7, in call_wilma\n')

    def testFormatException(self):
        try:
            bomb(AttributeError("barney"))
        except:
            lines = self.fmt.format_exception(*exc_info())
            self.assertEqual(lines[0], 'Traceback (most recent call last):\n')

            self.assertEqual(lines[1], '  File "%s", line 327, in testFormatException\n'
                                       '    bomb(AttributeError("barney"))\n' % __FILE__)

            self.assertEqual(lines[2], '  File "%s", line 26, in bomb\n'
                                       '    raise exception\n' % __FILE__)

            self.assertEqual(lines[3], 'AttributeError: barney\n')

    def testFormatExceptionWithTracebackInfo(self):
        try:
            bomb(AttributeError('barney'), tbi='wilma, betty')
        except:
            lines = self.fmt.format_exception(*exc_info())
            self.assertEqual(lines[2], '  File "%s", line 26, in bomb\n'
                                       '    raise exception\n'
                                       '   - __traceback_info__: wilma, betty\n' % __FILE__)

            self.assertEqual(lines[3], 'AttributeError: barney\n')

    def testFormatExceptionWithTracebackSupplement(self):
        try:
            bomb(AttributeError('barney'), tbs=(Supplement, 'http://bedrock.com', 23, 42))
        except:
            lines = self.fmt.format_exception(*exc_info())
            self.assertEqual(lines[2], '  File "%s", line 26, in bomb\n'
                                       '    raise exception\n'
                                       '   - http://bedrock.com\n'
                                       '   - Line 23, Column 42\n' % __FILE__)

            self.assertEqual(lines[3], 'AttributeError: barney\n')

    def testFormatExc(self):
        try:
            bomb(AttributeError("barney"))
        except:
            formatted = self.fmt.format_exc()
            self.assertEqual(formatted, 'Traceback (most recent call last):\n'
                                        '  File "%s", line 365, in testFormatExc\n'
                                        '    bomb(AttributeError("barney"))\n'
                                        '  File "%s", line 26, in bomb\n'
                                        '    raise exception\n'
                                        'AttributeError: barney\n' % (__FILE__, __FILE__))

    def testFormatExcWithTracebackInfo(self):
        try:
            bomb(AttributeError("barney"), tbi="wilma, betty")
        except:
            formatted = self.fmt.format_exc()
            self.assertEqual(formatted, 'Traceback (most recent call last):\n'
                                        '  File "%s", line 377, in testFormatExcWithTracebackInfo\n'
                                        '    bomb(AttributeError("barney"), tbi="wilma, betty")\n'
                                        '  File "%s", line 26, in bomb\n'
                                        '    raise exception\n'
                                        '   - __traceback_info__: wilma, betty\n'
                                        'AttributeError: barney\n' % (__FILE__, __FILE__))

    def testFormatExcWithTracebackSupplement(self):
        try:
            bomb(AttributeError("barney"), tbs=(Supplement, "http://bedrock.com", 23, 42))
        except:
            formatted = self.fmt.format_exc()
            self.assertEqual(formatted, 'Traceback (most recent call last):\n'
                                        '  File "%s", line 390, in testFormatExcWithTracebackSupplement\n'
                                        '    bomb(AttributeError("barney"), tbs=(Supplement, "http://bedrock.com", 23, 42))\n'
                                        '  File "%s", line 26, in bomb\n'
                                        '    raise exception\n'
                                        '   - http://bedrock.com\n'
                                        '   - Line 23, Column 42\n'
                                        'AttributeError: barney\n' % (__FILE__, __FILE__))

    def testGetTraceback(self):
        try:
            bomb(AttributeError("barney"), tbs=(Supplement, "http://bedrock.com", 23, 42))
        except:
            from django.contrib.exceptionformatter.installer import _get_traceback
            formatted = _get_traceback(None)
            self.assertEqual(formatted, 'Traceback (most recent call last):\n'
                                        '\n'
                                        '  File "%s", line 404, in testGetTraceback\n'
                                        '    bomb(AttributeError("barney"), tbs=(Supplement, "http://bedrock.com", 23, 42))\n'
                                        '\n'
                                        '  File "%s", line 26, in bomb\n'
                                        '    raise exception\n'
                                        '   - http://bedrock.com\n'
                                        '   - Line 23, Column 42\n'
                                        '\n'
                                        'AttributeError: barney\n' % (__FILE__, __FILE__))


def test_suite():
    return TestSuite((
        makeSuite(FormatterTests),
    ))


if __name__ == '__main__':
    main(defaultTest='test_suite')

